/*
 *  linux/fs/beos/util.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 */

#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/nls.h>

#define BEOS_UNKNOWN_CHAR '?'


/*
 * UTF-8 to NLS charset  convert routine
 */
char * beos_utf2nls (char * src, int srclen, char ** dist,
		int * len, struct super_block * sb)
{
	char *           	out;
	struct nls_table * 	nls = sb->u.beos_sb.nls;
	char *			orig = src;
	int			olen = 0;

	BEOS_OUTPUT (("---> utf2nls()\n"));

	*dist = out = (char *) __getname();
	
	if (!out)
	{
		printk(KERN_ERR "BEOS utf2nls: cannot allocate memory\n");
		return NULL;
	}

	if (!nls)
	{
		printk(KERN_INFO "BEOS utf2nls: '%s' not nls\n", src);
		while (*src && srclen-- > 0) { *out++ = *src++; }
	}
	else
	{
		printk(KERN_INFO "BEOS utf2nls: processing nls\n");
		while (*src && srclen > 0)
		{
			int n;
			unsigned char		page;
			__u16			w;

			/*
			 *  convert from UTF-8 to Unicode
			 */
			// int utf8_mbtowc(__u16 *, const __u8 *, int)
			n = utf8_mbtowc(&w, src, srclen);
			if (n <= 0)
			{
				/*
				 * cannot convert from UTF-8 to unicode
				 */
				*out++ = BEOS_UNKNOWN_CHAR;
				src++;
				srclen--;
				continue;
			}
			//else
			//{
			//	printk(KERN_INFO "BEOS utf8_mbtowc: %d\n",w);
			//}
			src += n;
			srclen -= n;

		//	printk(KERN_INFO "BEOS uni2char: working on %d\n",w);
			nls->uni2char(0x00, w, &page, 20, &olen);
		//	printk(KERN_INFO "BEOS uni2char: returned %d\n",page);

			if (!page)
				*out++ = BEOS_UNKNOWN_CHAR;
			else
				//*out++ = page[w & 0xff];
				*out++ = page;
		}
	}
	*out = '\0';
	*len = out - (*dist);

	BEOS_OUTPUT (("<--- utf2nls()\n"));
	if (nls)
	{
		printk(KERN_INFO "BEOS utf2nls in '%s'\tout '%s' length '%d'\n", orig, *dist, *len);
	}
	return *dist;
}


char * beos_nls2utf (unsigned char * src, int srclen, char ** dist,
		int * len, struct super_block * sb)
{
	char *			out;
	struct nls_table *	nls = sb->u.beos_sb.nls;
	char *			orig = src;

	//BEOS_OUTPUT (
	printk(KERN_INFO "BEOS ---> nls2utf()\n");

	*dist = out = (char *) __getname();
	if (!out)
	{
		//BEOS_OUTPUT (
		printk(KERN_WARNING "BEOS: cannot allocate memory\n");
		return NULL;
	}

	if (!nls)
	{
		printk(KERN_INFO "BEOS nls2utf: not nls\n");
		while (*src && srclen-- > 0) { *out++ = *src++; }
	}
	else
	{
		printk(KERN_INFO "BEOS nls2utf: processing nls\n");
		while (*src && srclen > 0)
		{
			int 		n,offset=0;
			__u16		w=0;
			__u8		wl,wh;

			/*
			 * convert from nls to unicode
			 */
		//	printk(KERN_INFO "BEOS char2uni: working on %d\n",*src);
			nls->char2uni(src, &offset, &wh, &wl);
		//	printk(KERN_INFO "BEOS char2uni: returned %d\n",wh);

			src++;
			srclen--;
			//w = wh >> 8 << wl;
			w = wh;
			//printk(KERN_INFO "BEOS char2uni: %d\n",w);
			/*
			 * convert from unicode to UTF-8
			 */
			n = utf8_wctomb(out, w, 6);
			if (n <= 0)
			{
				*out++ = BEOS_UNKNOWN_CHAR;
				continue;
			}
			out += n;
		}
	}

	*out = '\0';
	*len = out - (*dist);

	//BEOS_OUTPUT (
	printk(KERN_INFO "BEOS <--- nls2utf()\n");
	if (nls)
	{
		printk(
		KERN_INFO "BEOS nls2utf in '%s'\tout '%s' length '%d'\n", orig, *dist,*len);
	}
	return *dist;
}
